#include "fastml.h"
// TO DO
//1. ADD A CALL TO ROUTINE FOR ML BRANCH LENGTH OPTIMIZATION WITH AND WITHOUT GAMMA.

#include "aaJC.h"
#include "amino.h"
#include "chebyshevAccelerator.h"
#include "checkcovFanctors.h"
#include "clustalFormat.h"
#include "computeDownAlg.h"
#include "computeMarginalAlg.h"
#include "computeUpAlg.h"
#include "distanceTable.h"
#include "errorMsg.h"
#include "fastaFormat.h"
#include "fastmlCompute.h"
#include "jcDistance.h"
#include "likeDist.h"
#include "maseFormat.h"
#include "molphyFormat.h"
#include "nj.h"
#include "nucJC.h"
#include "nucleotide.h"
#include "numRec.h"
#include "phylipFormat.h"
#include "readDatMatrix.h"
#include "sequenceContainer1G.h"
#include "someUtil.h"
#include "trivialAccelerator.h"
#include "uniDistribution.h"

#include <iostream>
#include <cassert>
#include <algorithm>
using namespace std;



fastml::fastml(int argc, char* argv[]) {

	fillOptionsParameters(argc,argv);
	printFASTMLInfo(cout);
	printOptionParameters();
	getStartingSequenceData();
	if (_sc.numberOfSequences()==0) errorMsg::reportError("zero number of sequences in sequence file");
	getStartingStochasticProcess();// must be b4 the tree!
	getStartingEvolTree();
	outputTreeFileInAncestorFormatToAFile();//DEFAULT = FIN
	outputTreeFileInPHYLIPFormatToAFile();//DEFAULT = PIN
	computeJointReconstruction();
	computeMarginalReconstruction();
}

void fastml::printFASTMLInfo(ostream& out) {
	out<<endl;
	out<<" ======================================================="<<endl;
	out<<" FASTML: Fast ancestral sequence reconstruction using ML"<<endl;
	out<<" Tal Pupko: talp@post.tau.ac.il                         "<<endl;
	out<<" ======================================================="<<endl;
	out<<endl;
}

void fastml::fillOptionsParameters(int argc, char* argv[]) {
	_options = new fastmlOptions(argc, argv);
}

void fastml::printOptionParameters() {
	cout<<"\n ---------------------- THE PARAMETERS ----------------------------"<<endl;
	if (_options->treefile.size()>0) cout<<"tree file is : "<<_options->treefile<<endl;
	if (_options->seqfile.size()>0) cout<<"seq file is : "<<_options->seqfile<<endl;
	if (_options->outFile_fin.size()>0) cout<<"FASTML format file with internal nodes is : "<<_options->outFile_fin<<endl;
	if (_options->outFile_jpf.size()>0) cout<<"Output file of joint probabilities at each position is : "<<_options->outFile_jpf<<endl;
	if (_options->outFile_jrof.size()>0) cout<<"Output file of the joint reconstruction is : "<<_options->outFile_jrof<<endl;
	if (_options->outFile_mrof.size()>0) cout<<"Output file of the marginal reconstruction is : "<<_options->outFile_mrof<<endl;
	if (_options->outFile_pin.size()>0) cout<<"Phylip tree file output file with internal nodes is : "<<_options->outFile_pin<<endl;
	if (_options->outFile_mdp.size()>0) cout<<"Output file of the marginal probabilities at each site at each node: "<<_options->outFile_mdp<<endl;
	switch (_options->modelName){
		case (fastmlOptions::day): cout<< "probablistic_model = DAY" <<endl; break;
		case (fastmlOptions::jtt): cout<< "probablistic_model = JTT" <<endl; break;
		case (fastmlOptions::rev): cout<< "probablistic_model = REV" <<endl; break;
		case (fastmlOptions::aajc): cout<< "probablistic_model = AAJC" <<endl; break;
		case (fastmlOptions::nucjc): cout<< "probablistic_model = NUCJC" <<endl; break;
	}

	if (_options->removeGaps) {
		cout<<" positions with gaps were removed from the analysis."<<endl;
		cout<<" the number of each position in the results refers to the gapless alignment."<<endl;
	}
	else {
		cout<<"(gaps characters were treated as missing data.)"<<endl;
	}

	cout<<"\n -----------------------------------------------------------------"<<endl;
}

void fastml::getStartingSequenceData(){
	if (_options->seqfile == "") {
		errorMsg::reportError("Please give a sequence file name in the command line");
	}
	ifstream in(_options->seqfile.c_str());
	int alphabetSize = _options->alphabet_size;
	if (alphabetSize==4) _alph = new nucleotide;
	else if (alphabetSize == 20) _alph = new amino;
	else errorMsg::reportError("no such alphabet in function fastml::getStartingSequenceData");

	sequenceContainer1G original;
	switch (_options->seqInputFormat) {
	case fastmlOptions::mase: 
			original= maseFormat::read(in,_alph);
			break;
	case fastmlOptions::molphy: 
			original=molphyFormat::read(in,_alph);
			break;
	case fastmlOptions::clustal: 
			original=clustalFormat::read(in,_alph);
			break;
	case fastmlOptions::fasta: 
			original=fastaFormat::read(in,_alph);
			break;
	case fastmlOptions::phylip: 
			original=phylipFormat::read(in,_alph);
			break;
	default: errorMsg::reportError(" format not implemented yet in this version... ",1);
	}

	if (_options->removeGaps) {
//		vector<int> parCol;
//		original.getParticiantColVecAccordingToGapCols(parCol);
//		_sc = new sequenceData(original,parCol);
		errorMsg::reportError("removeGaps currently not implemented");
	} else {
		original.changeGapsToMissingData();
		_sc = original;
	}
}

//	void makeSureTreeIsBiFurcatingAndUnrooted();
void fastml::getStartingStochasticProcess() {
	distribution *dist = NULL;
	switch (_options->distributionName){
	case (fastmlOptions::hom): dist =  new uniDistribution; break;
		default: errorMsg::reportError("this distribution name is not yet available");

	}

	replacementModel *probMod=NULL;
	pijAccelerator *pijAcc=NULL;
	switch (_options->modelName){
		case (fastmlOptions::day):
			probMod=new pupAll(datMatrixHolder::dayhoff);pijAcc = new chebyshevAccelerator(probMod); break;
		case (fastmlOptions::jtt):
			probMod=new pupAll(datMatrixHolder::jones); pijAcc = new chebyshevAccelerator(probMod); break;
		case (fastmlOptions::rev):
			probMod=new pupAll(datMatrixHolder::mtREV24); pijAcc = new chebyshevAccelerator(probMod); break;
		case (fastmlOptions::wag):
			probMod=new pupAll(datMatrixHolder::wag); pijAcc = new chebyshevAccelerator(probMod); break;
		case (fastmlOptions::cprev):
			probMod=new pupAll(datMatrixHolder::cpREV45); pijAcc = new chebyshevAccelerator(probMod); break;
		case (fastmlOptions::nucjc):
			probMod=new nucJC; pijAcc = new trivialAccelerator(probMod); break;
		case (fastmlOptions::aajc):
			probMod=new aaJC; pijAcc = new trivialAccelerator(probMod); break;
		default:
			errorMsg::reportError("this probablistic model is not yet available");
	}
	_sp = new stochasticProcess(dist, pijAcc);


	if (probMod) delete probMod;
	if (pijAcc) delete pijAcc;
	if (dist) delete dist;}



void fastml::getStartingEvolTree(){
	if (strcmp(_options->treefile.c_str(),"")==0) {
		switch (_options->treeSearchAlg){
			case (fastmlOptions::njJC):
				getStartingNJtreeNjJC(); 
				break;
			/*case (rate4siteOptions::njJCOLD):
				getStartingNJtreeNjJC_old(); 
				break;*/
			/*case (rate4siteOptions::njML): {
					getStartingNJtreeNjMLdis();
				}*/
				//break;
			//case (rate4siteOptions::MLfromManyNJ): {
				//cerr<<"computing the ML tree... "<<endl;
				//getStartingMLtreeFromManyNJtrees();
				//}break;
			default:
				errorMsg::reportError("This tree search mode is not yet available");
		}
	}
	else getStartingTreeFromTreeFile();
}

void fastml::getStartingTreeFromTreeFile(){
	_t1= tree(_options->treefile);
	if (!_t1.WithBranchLength()) {
		errorMsg::reportError("input tree is without branch lengths. Branch lengths are needed for ancestral sequence reconstruction\n");
	}
}

void fastml::outputTreeFileInAncestorFormatToAFile() {
	ofstream outputInAncestorFileFormat(_options->outFile_fin.c_str());
	_t1.output(outputInAncestorFileFormat,tree::ANCESTOR,true);
	outputInAncestorFileFormat.close();
}

void fastml::outputTreeFileInPHYLIPFormatToAFile() {
	ofstream outputInPhylipFileFormat(_options->outFile_pin.c_str());
	_t1.output(outputInPhylipFileFormat,tree::PHYLIP,true);
	outputInPhylipFileFormat.close();
}

void fastml::computeJointReconstruction() {
	computePijHom pi;
	suffStatGlobalHomPos ssc;
	suffStatGlobalHomPosFASTML sscFASTML;
	ssc.allocatePlace(_t1.iNodes(),_sc.alphabetSize());
	sscFASTML.allocatePlace(_t1.iNodes(),_sc.alphabetSize());

	pi.fillPij(_t1,*_sp,_sc.alphabetSize());
	vector<string> ancestralSequences(_t1.iNodes());
	MDOUBLE totalLikelihoodOfReconstruction = 0;
	vector<MDOUBLE> jointLikelihoodOfPositions;
	for (int pos=0; pos<_sc.seqLen(); ++pos) {
		cerr<<"doing position (joint): "<<pos<<endl;
		fastmlCompute fastmlCompute1;
		fastmlCompute1.fillComputeUp(_t1,_sc,pos,pi,ssc,sscFASTML);
		MDOUBLE likelihoodOfPos = 0;

		vector<int> res =fastmlCompute1.computeJointAncestralFromSSC(_t1,_sc,*_sp,pos,pi,ssc,sscFASTML,likelihoodOfPos);
		treeIterDownTopConst tIt(_t1);
		for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
			if (treeInterface::isInternal(mynode)) { 
				ancestralSequences[mynode->id()]+=_sc.getAlphabet()->fromInt(res[mynode->id()]);
			}
		}
		jointLikelihoodOfPositions.push_back(likelihoodOfPos);
	}

	outputTheJointProbAtEachSite(jointLikelihoodOfPositions);//(DEFAULT = JPF)
	sequenceContainer1G scJoint = getJointReconstruction(ancestralSequences);
	outputTheJointReconstruction(scJoint);//(DEFAULT = JROF).
}

void fastml::outputTheJointProbAtEachSite(const vector<MDOUBLE> & jointLikelihoodOfPositions) {
	ofstream jointProbOutput(_options->outFile_jpf.c_str());
	MDOUBLE totalLogLikelihood =0;
	for (int j=0; j < jointLikelihoodOfPositions.size(); ++j) {
		totalLogLikelihood+=log(jointLikelihoodOfPositions[j]);
		jointProbOutput<<"Joint probability of position "<<j<<": "<<jointLikelihoodOfPositions[j]<<endl;
	}
	jointProbOutput<<"total log likelihood of joint reconstruction: "<<totalLogLikelihood<<endl;
	jointProbOutput.close();
}

sequenceContainer1G fastml::getJointReconstruction(const vector<string> & ancestralSequences){
	sequenceContainer1G withAncestral = _sc;
	treeIterDownTopConst tIt2(_t1);
	for (tree::nodeP mynode = tIt2.first(); mynode != tIt2.end(); mynode = tIt2.next()) {
		if (treeInterface::isInternal(mynode)) { 
			sequence tmp(ancestralSequences[mynode->id()],mynode->name(),"joint reconstruction",mynode->id(),_sc.getAlphabet());
			withAncestral.add(tmp);
		}
	}
	return withAncestral;
}


void fastml::outputTheJointReconstruction(const sequenceContainer1G& withAncestral){
	ofstream jointReconstructionOutputFile(_options->outFile_jrof.c_str());
	switch (_options->seqOutputFormat) {
	case fastmlOptions::mase: 
			 maseFormat::write(jointReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::molphy: 
			molphyFormat::write(jointReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::clustal: 
			clustalFormat::write(jointReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::fasta: 
			fastaFormat::write(jointReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::phylip: 
			phylipFormat::write(jointReconstructionOutputFile,withAncestral);
			break;
	default: errorMsg::reportError(" format not implemented yet in this version... ",1);
	}
	jointReconstructionOutputFile.close();
}

void fastml::computeMarginalReconstruction(){
	computePijHom pi;
	pi.fillPij(_t1,*_sp,_sc.alphabetSize());
	vector<string> ancestralSequences(_t1.iNodes());
	fastmlCompute fastmlCompute1;
	MDOUBLE totalLikelihoodOfReconstruction = 0;
	vector<MDOUBLE> marginalLikelihoodOfPositions;
	for (int pos=0; pos<_sc.seqLen(); ++pos) {
		suffStatGlobalHomPos sscUp;// this is for a specific position.
		suffStatGlobalHomPos sscDown;// this is for a specific position.
		suffStatGlobalHomPos sscMarginal; // this is for a specific position.
		sscUp.allocatePlace(_t1.iNodes(),_sc.alphabetSize());
		sscDown.allocatePlace(_t1.iNodes(),_sc.alphabetSize());
		sscMarginal.allocatePlace(_t1.iNodes(),_sc.alphabetSize());

		cerr<<"doing position (marginal): "<<pos<<endl;
		computeUpAlg computeUpAlg1;
		computeUpAlg1.fillComputeUp(_t1,_sc,pos,*_sp,sscUp);

		computeDownAlg computeDownAlg1;
		computeDownAlg1.fillComputeDown(_t1,_sc,pos,pi,sscDown,sscUp);

		computeMarginalAlg computeMarginalAlg1;
		MDOUBLE posProb =0;
		computeMarginalAlg1.fillComputeMarginal(_t1,_sc,*_sp,pos,pi,sscMarginal,sscUp,sscDown,posProb);
		MDOUBLE likelihoodOfPos = 0;

		vector<int> res =fastmlCompute1.computeMarginalAncestralFromSSC(sscMarginal,*_sp,_t1);
		treeIterDownTopConst tIt(_t1);
		for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
			if (treeInterface::isInternal(mynode)) { 
				ancestralSequences[mynode->id()]+=_sc.getAlphabet()->fromInt(res[mynode->id()]);
			}
		}
		marginalLikelihoodOfPositions.push_back(posProb);
		outputTheMarginalProbForEachCharForEachNode(sscMarginal,pos);//(DEFAULT = JPF, same file as above).
	}

	outputTheMarginalProbAtEachSite(marginalLikelihoodOfPositions);//(DEFAULT = JPF)
	sequenceContainer1G scMarginal = getMarginalReconstruction(ancestralSequences);
	outputTheMarginalReconstruction(scMarginal);//(DEFAULT = ).

}

void fastml::outputTheMarginalProbAtEachSite(const vector<MDOUBLE> &  marginalLikelihoodOfPositions){//(DEFAULT = MROF)
	ofstream marginalProbOutput(_options->outFile_mdp.c_str(),ios::app);
	marginalProbOutput<<"\nThe marginal probabilities at each site: \n";
	MDOUBLE totalLogLikelihood =0;
	for (int j=0; j < marginalLikelihoodOfPositions.size(); ++j) {
		totalLogLikelihood+=log(marginalLikelihoodOfPositions[j]);
		marginalProbOutput<<"Marginal probability of position "<<j<<": "<<marginalLikelihoodOfPositions[j]<<endl;
		
	}
	marginalProbOutput<<"total log likelihood of marginal reconstruction: "<<totalLogLikelihood<<endl;
	marginalProbOutput.close();
}

void fastml::outputTheMarginalProbForEachCharForEachNode(const suffStatGlobalHomPos & sscMarginal,
														 const int pos){//(DEFAULT = JPF, same file as above).
	if (pos==0) {ofstream x(_options->outFile_mdp.c_str()); x.close();} // initializing a new file for the first position.
	ofstream marginalProbOutput(_options->outFile_mdp.c_str(),ios::app);
	treeIterDownTopConst tIt(_t1);
	marginalProbOutput<<"marginal probabilities at positions: "<<pos<<endl;
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
		if (treeInterface::isLeaf(mynode)) continue;
		marginalProbOutput<<"of node: "<<mynode->name()<<": ";
		vector<pair< MDOUBLE,string> > pres;
		int c=0;
		for (c=0; c < _sp->alphabetSize(); ++c) {
			pres.push_back(pair<MDOUBLE,string>(sscMarginal.get(mynode->id(),c),_sc.getAlphabet()->fromInt(c)));
		}
		sort(pres.begin(),pres.end());
		for (c=pres.size()-1; c >=0 ; --c) {
			if (pres[c].first<0.0001) continue;
			marginalProbOutput<<"p("<<pres[c].second;
			marginalProbOutput<<")="<<pres[c].first<<" ";
		}






		//for (c=0; c < _sp->alphabetSize(); ++c) {
		// here we output the probability of c in position pos in node mynode.
			//if (sscMarginal.get(mynode->id(),c)<0.0001) continue;
			//marginalProbOutput<<"p("<<_sc.getAlphabet()->fromInt(c)<<")="<<sscMarginal.get(mynode->id(),c)<<" ";
		//}
		marginalProbOutput<<endl;
	}
	marginalProbOutput<<endl;
	marginalProbOutput.close();
}

sequenceContainer1G fastml::getMarginalReconstruction(const vector<string> & ancestralSequences){
	sequenceContainer1G withAncestral = _sc;
	treeIterDownTopConst tIt2(_t1);
	for (tree::nodeP mynode = tIt2.first(); mynode != tIt2.end(); mynode = tIt2.next()) {
		if (treeInterface::isInternal(mynode)) { 
			sequence tmp(ancestralSequences[mynode->id()],mynode->name(),"marginal reconstruction",mynode->id(),_sc.getAlphabet());
			withAncestral.add(tmp);
		}
	}
	return withAncestral;
}

void fastml::outputTheMarginalReconstruction(const sequenceContainer1G& withAncestral){
	ofstream marginalReconstructionOutputFile(_options->outFile_mrof.c_str());
	switch (_options->seqOutputFormat) {
	case fastmlOptions::mase: 
			 maseFormat::write(marginalReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::molphy: 
			molphyFormat::write(marginalReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::clustal: 
			clustalFormat::write(marginalReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::fasta: 
			fastaFormat::write(marginalReconstructionOutputFile,withAncestral);
			break;
	case fastmlOptions::phylip: 
			phylipFormat::write(marginalReconstructionOutputFile,withAncestral);
			break;
	default: errorMsg::reportError(" format not implemented yet in this version... ",1);
	}
	marginalReconstructionOutputFile.close();
}

void fastml::getStartingNJtreeNjJC() {
	jcDistance pd1(_options->alphabet_size);
	VVdouble disTab;
	vector<string> vNames;
	giveDistanceTable(&pd1,
					   _sc,
					   disTab,
					   vNames);
	getStartingTreeNJ_fromDistances(disTab,vNames);
} 

void fastml::getStartingTreeNJ_fromDistances(const VVdouble& disTab,
	const vector<string>& vNames) {
	NJalg nj1;
	_t1= nj1.computeNJtree(disTab,vNames);
	//ofstream f;
	//string fileName1=_options->treeOutFile;
	//f.open(fileName1.c_str());
	//_t1.output(f);
	//f.close();
	//cout<<"The tree was written to a file name called "<<_options->treeOutFile<<endl;
}