#include "computeUpAlg.h"
#include "definitions.h"
#include "seqContainerTreeMap.h"

#include "suffStatComponentFASTML.h"
#include "fastmlCompute.h"


#include <iostream>
#include <cassert>
using namespace std;

void fastmlCompute::fillComputeUp(const tree& et,
				   const sequenceContainer& sc,
				   const int pos,
				   const computePijHom& pi,
				   suffStatGlobalHomPos& ssc,
				   suffStatGlobalHomPosFASTML& sscFASTML) {

	seqContainerTreeMap sctm(sc,et);
	ssc.allocatePlace(et.iNodes(),pi.alphabetSize());
	treeIterDownTopConst tIt(et);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
			if (mynode->sons.empty()) {// leaf
			for(int letterInFather=0; letterInFather<pi.alphabetSize();letterInFather++) {
				const int seqID = sctm.seqIdOfNodeI(mynode->id());
				MDOUBLE totalVal = 0.0;
				for (int let=0; let<pi.alphabetSize();let++) {
					MDOUBLE val = sc.getAlphabet(pos)->relations(sc[seqID][pos],let);
					if (val>0) {
						val*=pi.getPij(mynode->id(),letterInFather,let);
						totalVal +=val;
					}
				}
				//cerr<<"val =" << val <<" "; // REMOVE!
				//cerr<<"_pi->data(mynode->id(),pos)= "<<_pi->data(mynode->id(),pos)<<" ";//REMOVE
				ssc.set(mynode->id(),letterInFather,totalVal);
				sscFASTML.set(mynode->id(),letterInFather,sc[seqID][pos]);
			}
		}
		else {
			for(int letterInFather=0; letterInFather<pi.alphabetSize();letterInFather++) {
				MDOUBLE maxProb=0.0;
				int bestLet = -1;
				for (int let=0; let<pi.alphabetSize();++let) {
					MDOUBLE tmpProb = 1;
					if (mynode->isRoot() == false) {
						tmpProb *= pi.getPij(mynode->id(),letterInFather,let);
					}
					for(int i=0; i < mynode->sons.size();++i){				
						tmpProb *= ssc.get(mynode->sons[i]->id(),let);
					}
					if (tmpProb>maxProb) {
						maxProb = tmpProb;
						bestLet = let;
					}
				}
				ssc.set(mynode->id(),letterInFather,maxProb);
				sscFASTML.set(mynode->id(),letterInFather,bestLet);
				if (mynode->isRoot()) break; // there's no meening to letterInFather in case of root.
			}
		}
	}
}

vector<int> fastmlCompute::computeJointAncestralFromSSC(const tree& et,
				   const sequenceContainer& sc,
				   const stochasticProcess& sp,
				   const int pos,
				   const computePijHom& pi,
				   const suffStatGlobalHomPos& ssc,
				   const suffStatGlobalHomPosFASTML& sscFASTML,
				   MDOUBLE & likelihoodOfReconstruction) {
	treeIterTopDownConst tIt(et);
	vector<int> res(et.iNodes());
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
		if (mynode->isRoot() == false) {
			int letterInFather = res[mynode->father()->id()];
			int tmp = sscFASTML.get(mynode->id(),letterInFather);
			res[mynode->id()] = tmp;
		} else {//special case of the root
			MDOUBLE maxL = VERYSMALL;
			int bestCharInRoot = sscFASTML.get(mynode->id(),0);
			likelihoodOfReconstruction = ssc.get(mynode->id(),0)*sp.freq(bestCharInRoot);;
			res[mynode->id()] = bestCharInRoot;
		}
	}
	return res;
}

vector<int> fastmlCompute::computeMarginalAncestralFromSSC(
	const suffStatGlobalHomPos& ssc,
	const stochasticProcess & sp,
	const tree& et){
	treeIterTopDownConst tIt(et);
	vector<int> res(et.iNodes());
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
		MDOUBLE bestP =-1.0;
		int bestChar = -1;
		
		for (int i=0; i < sp.alphabetSize(); ++i) {
			if (ssc.get(mynode->id(),i) > bestP) {
				bestP = ssc.get(mynode->id(),i);
				bestChar = i;
			}
		}
		res[mynode->id()] = bestChar;
	}
	return res;
}









