#include "fastml_options.h"
#include "errorMsg.h"
#include <iostream>
using namespace std;

fastmlOptions::fastmlOptions(int& argc, char *argv[]):

// DEFAULTS VALUES:

// sequence input format information
seqfile(""),
seqInputFormat(phylip),

// tree input format information
treefile(""),

// log file format information (for debuggin only)
logFile(""), //log.txt
logValue(3),

// output file information
seqOutputFormat(phylip),
outFile_jrof("jrof.txt"), //(-j) joint reconstruction output file
outFile_mrof("mrof.txt"), //(-k) marginal reconstruction output file
outFile_pin("pin.txt"), // (-p) phylip tree file format with internal nodes lables
outFile_fin("fin.txt"), // (-f) FASTML tree file format with internal nodes lables
outFile_jpf("jpf.txt"), // (-e) joint probability of each position file
outFile_mdp("mpf.txt"), // (-d) marginal probabilities of each position, and at each node file

// bbl options
bblOption(non),
// model:
modelName(jtt),
alphabet_size(20), // this is 20
distributionName(hom),
treeSearchAlg(njJC),
removeGaps(false), // default gaps = missing data.
outPtr(&cout) {


	//static struct option long_options[] =  {{0, 0, 0, 0}};
	int option_index = 0;
	int c=0;
	bool algo_set=false;

//	out_f.open(outFile.c_str()); // default output file
//	outPtr=&out_f; // default output file

	while (c >= 0) {
//		c = getopt_long(argc, argv,"p:P:z:Z:bBm:M:T:t:s:S:l:L:v:V:Q:q:DJR:r:C:djc:Hh?",NULL,&option_index);
//			c = getopt(argc, argv,"z:Z:A:a:bBm:M:T:t:s:S:O:o:l:L:v:V:Q:q:DJRC:djrc:Hh?");
		c = getopt_long(argc, argv,"d:D:e:E:f:F:Hhj:J:k:K:m:M:p:P:q:Q:r:R:s:S:t:T:",NULL,&option_index);

		switch (c) {

			// tree file, seqfile and sequence input file format.
			//case 'b':case 'B': removeGaps=true; break;
			case 'd':case 'D': outFile_mdp=optarg; break;
			case 'e':case 'E': outFile_jpf=optarg; break;
			case 'f':case 'F': outFile_fin=optarg; break;

			case 'h':case 'H': case '?':
				cout <<"USAGE:	"<<argv[0]<<" [-options] "<<endl <<endl;
				cout <<"|/-------------------- +-----------------------+"<<endl;
				cout <<"|-t    tree file       |-Q    input seq format |"<<endl;
				cout <<"|-s    seq file        |-Qc   Clustal          |"<<endl;
				cout <<"|                      |-Qf   Fasta            |"<<endl;
				cout <<"|                      |-Qm   Mase             |"<<endl;
				cout <<"|                      |-Qmo  Molphy           |"<<endl;
				cout <<"|                      |-Qp   Phylip           |"<<endl;
				cout <<"|/-------------------- +-----------------------+"<<endl;
				cout <<"|-j joint output file  |-R    output seq format|"<<endl;
				cout <<"|-k marginal out file  |-Rc   Clustal          |"<<endl;
				cout <<"|-P phylip tree file   |-Rf   Fasta            |"<<endl;
				cout <<"| (with internal nodes)|-Rm   Mase             |"<<endl;
				cout <<"|-f FASTML tree file   |-Rmo  Molphy           |"<<endl;
				cout <<"|-e joint probabilities|-Rp   Phylip           |"<<endl;
				cout <<"|-d marginal probabilities                     |"<<endl;
				cout <<"|----------------------|-----------------------|"<<endl;
				cout <<"|-M     model name                             |"<<endl;
				cout <<"|-Mj    JTT                                    |"<<endl;
				cout <<"|-Mr    REV (for mitochondrial genomes)        |"<<endl;
				cout <<"|-Md    DAY                                    |"<<endl;
				cout <<"|-Mw    WAG                                    |"<<endl;
				cout <<"|-MC    cpREV (for chloroplasts genomes)       |"<<endl;
				cout <<"|-Ma    JC amino acids                         |"<<endl;
				cout <<"|-Mn    JC nucleotides                         |"<<endl;
				cout <<"|----------------------------------------------|"<<endl;
				//cout <<"|-b remove positions with gaps                 |"<<endl;
				//cout <<"|default: gaps are treated as missing data     |"<<endl;
				//cout <<"|----------------------------------------------|"<<endl;
				cout <<"|-h or -? or -H     help                       |"<<endl;
				cout <<"|capital and no captial letters are ok         |"<<endl;
				cout <<"+----------------------+-----------------------+"<<endl;
				//cout <<"|-z branch length optimization:                |"<<endl;
				//cout <<"|-za ML branch lengths                         |"<<endl;
				//cout <<"|-zb ML branch lengths with Gamma              |"<<endl;
				//cout <<"|----------------------------------------------|"<<endl;
				cout<<endl;	cerr<<" please press 0 to exit "; int d; cin>>d;exit (0);
			case 'j':case 'J': outFile_jrof=optarg; break;
			case 'k':case 'K': outFile_mrof=optarg; break;
			case 'p':case 'P': outFile_pin=optarg; break;

			
				//case 'l':case 'L': logFile=optarg; break;
			case 'm':case 'M':	{
				switch (optarg[0]) {
					case 'd': case 'D':  modelName=day;alphabet_size=20; break;
					case 'j': case 'J':  modelName=jtt;alphabet_size=20; break;
					case 'r': case 'R':  modelName=rev;alphabet_size=20; break;
					case 'w': case 'W':  modelName=wag;alphabet_size=20; break;
					case 'c': case 'C':  modelName=cprev;alphabet_size=20; break;
					case 'a': case 'A':  modelName=aajc;alphabet_size=20; break;
					case 'n': case 'N':  modelName=nucjc;alphabet_size=4; break;
					default:modelName=jtt;alphabet_size=20;
					break;
				}
			} break;

			case 'q':case 'Q':	{
				switch (optarg[0]) {
					case 'C': case 'c':  seqInputFormat=clustal; break;
					case 'F': case 'f':  seqInputFormat=fasta; break;
					case 'M': case 'm': switch (optarg[1]) {
						case 'o': case 'O': seqInputFormat=molphy; break;
						default: seqInputFormat=mase;	break;
					} break;
					case 'P': case 'p':  seqInputFormat=phylip; break;
					default:seqInputFormat=mase;
					break;
				} // end of switch (optarg[0]) {
				
			}break; // end of case 'Q':
			
			case 'r':case 'R':	{
				switch (optarg[0]) {
					case 'C': case 'c':  seqOutputFormat=clustal; break;
					case 'F': case 'f':  seqOutputFormat=fasta; break;
					case 'M': case 'm': switch (optarg[1]) {
						case 'o': case 'O': seqOutputFormat=molphy; break;
						default: seqOutputFormat=mase;	break;
					} break;
					case 'P': case 'p':  seqOutputFormat=phylip; break;
					default:seqOutputFormat=mase;
					break;
				} // end of switch (optarg[0]) {
				
			}break; // end of case 'Q':
			
			case 's':case 'S': seqfile=optarg; break;
			case 't':case 'T': treefile=optarg; break;
			//case 'v':case 'V': logValue=atoi(optarg); break;
			/*case 'z':case 'Z': {
				switch (optarg[0]) {
					case 'A': case 'a':  bblOption=ml; break;
					case 'B': case 'b':  bblOption=ml_gam; break;
					default:bblOption=non;
					break;
				}
			} break;*/

		}
	}
}

